import unittest
from translationstring import text_type
from translationstring.compat import u

class TestTranslationString(unittest.TestCase):
    def _getTargetClass(self):
        from translationstring import TranslationString
        return TranslationString
        
    def _makeOne(self, msgid, **kw):
        klass = self._getTargetClass()
        return klass(msgid, **kw)

    def test_is_text_type_subclass(self):
        inst = self._makeOne('msgid')
        self.assertTrue(isinstance(inst, text_type))

    def test_msgid_is_translation_string(self):
        another = self._makeOne('msgid', domain='domain', default='default',
                                mapping={'a':1})
        result = self._makeOne(another)
        self.assertEqual(result, 'msgid')
        self.assertEqual(result.domain, 'domain')
        self.assertEqual(result.default, 'default')
        self.assertEqual(result.mapping, {'a':1})

    def test_default_None(self):
        inst = self._makeOne('msgid')
        self.assertEqual(inst, 'msgid')
        self.assertEqual(inst.default, 'msgid')

    def test_default_not_None(self):
        inst = self._makeOne('msgid', default='default')
        self.assertEqual(inst, 'msgid')
        self.assertEqual(inst.default, 'default')

    def test_allargs(self):
        inst = self._makeOne('msgid', domain='domain', default='default',
                             mapping='mapping')
        self.assertEqual(inst, 'msgid')
        self.assertEqual(inst.default, 'default')
        self.assertEqual(inst.mapping, 'mapping')
        self.assertEqual(inst.domain, 'domain')

    def test_format_no_exising_mapping(self):
        inst = self._makeOne('msgid', domain='domain', default='default')
        mapping = {'foo': 'bar'}
        new_inst = inst % mapping
        self.assertTrue(new_inst is not inst)
        self.assertEqual(new_inst, 'msgid')
        self.assertEqual(new_inst.default, 'default')
        self.assertEqual(new_inst.domain, 'domain')
        self.assertEqual(new_inst.mapping, mapping)
        # Make sure a copy is made to prevent changes in the original
        # dictionary from modifying our mapping.
        self.assertTrue(new_inst.mapping is not mapping)

    def test_format_extend_existing_mapping(self):
        inst = self._makeOne('msgid', domain='domain', default='default',
                mapping={'one': '1', 'two': '3'})
        new_inst = inst % {'two': '2', 'three': '3'}
        self.assertEqual(
                new_inst.mapping,
                {'one': '1', 'two': '2', 'three': '3'})
        # Make sure original is not changed
        self.assertEqual(inst.mapping, {'one': '1', 'two': '3'})

    def test_format_required_dict_argument(self):
        import operator
        inst = self._makeOne('msgid', domain='domain', default='default')
        self.assertRaises(ValueError, operator.mod, inst, ('foo',))


    def test_interpolate_substitution(self):
        mapping = {"name": "Zope", "version": 3}
        inst = self._makeOne('This is $name version ${version}.',
                              mapping=mapping)
        result = inst.interpolate()
        self.assertEqual(result, u('This is Zope version 3.'))

    def test_interpolate_subsitution_more_than_once(self):
        mapping = {"name": "Zope", "version": 3}
        inst = self._makeOne(
            u("This is $name version ${version}. ${name} $version!"),
            mapping=mapping)
        result = inst.interpolate()
        self.assertEqual(result, u('This is Zope version 3. Zope 3!'))

    def test_interpolate_double_dollar_escape(self):
        mapping = {"name": "Zope", "version": 3}
        inst = self._makeOne('$$name', mapping=mapping)
        result = inst.interpolate()
        self.assertEqual(result, u('$$name'))

    def test_interpolate_missing_not_interpolated(self):
        mapping = {"name": "Zope", "version": 3}
        inst = self._makeOne(
            u("This is $name $version. $unknown $$name $${version}."),
            mapping=mapping)
        result = inst.interpolate()
        self.assertEqual(result,
                         u('This is Zope 3. $unknown $$name $${version}.'))

    def test_interpolate_missing_mapping(self):
        inst = self._makeOne(u("This is ${name}"))
        result = inst.interpolate()
        self.assertEqual(result, u('This is ${name}'))

    def test_interpolate_passed_translated(self):
        mapping = {"name": "Zope", "version": 3}
        inst = self._makeOne(u("This is ${name}"), mapping = mapping)
        result = inst.interpolate('That is ${name}')
        self.assertEqual(result, u('That is Zope'))

    def test___reduce__(self):
        klass = self._getTargetClass()
        inst = self._makeOne('msgid', default='default', domain='domain',
                             mapping='mapping')
        result = inst.__reduce__()
        self.assertEqual(result, (klass, (u('msgid'), 'domain', u('default'), 
                                          'mapping', None)))

    def test___getstate__(self):
        inst = self._makeOne('msgid', default='default', domain='domain',
                             mapping='mapping')
        result = inst.__getstate__()
        self.assertEqual(result,
                         (u('msgid'), 'domain', u('default'), 'mapping', None))

class TestTranslationStringFactory(unittest.TestCase):
    def _makeOne(self, domain):
        from translationstring import TranslationStringFactory
        return TranslationStringFactory(domain)

    def test_allargs(self):
        factory = self._makeOne('budge')
        inst = factory('msgid', mapping='mapping', default='default')
        self.assertEqual(inst, 'msgid')
        self.assertEqual(inst.domain, 'budge')
        self.assertEqual(inst.mapping, 'mapping')
        self.assertEqual(inst.default, 'default')

    def test_msgid_is_translation_string_override_domain(self):
        user_factory = self._makeOne('user')
        factory = self._makeOne('budge')

        wrapped_inst = user_factory('wrapped_msgid', mapping={'a':1}, default='default')
        wrapper_inst = factory(wrapped_inst)

        self.assertEqual(str(wrapper_inst), 'wrapped_msgid')
        self.assertEqual(wrapper_inst.domain, 'user')

    def test_msgid_is_translation_string_override_kwarg(self):
        user_factory = self._makeOne('user')
        factory = self._makeOne('budge')

        wrapped_inst = user_factory('wrapped_msgid', mapping={'a':1}, default='default')
        wrapper_inst = factory(wrapped_inst, mapping={'b':1}, default='other_default')

        self.assertEqual(str(wrapper_inst), 'wrapped_msgid')
        self.assertEqual(wrapper_inst.mapping, {'b':1})
        self.assertEqual(wrapper_inst.default, 'other_default')


class TestChameleonTranslate(unittest.TestCase):
    def _makeOne(self, translator):
        from translationstring import ChameleonTranslate
        return ChameleonTranslate(translator)

    def test_msgid_nonstring(self):
        translate = self._makeOne(None)
        result = translate(None)
        self.assertEqual(result, None)

    def test_msgid_not_none_not_string(self):
        translate = self._makeOne(None)
        result = translate(True)
        self.assertEqual(result, 'True')

    def test_chameleon_default_marker_returned(self):
        # Chameleon uses a special StringMarker class as default value so it
        # can detect missing translations.
        class StringMarker(str): pass
        translate = self._makeOne(None)
        marker = StringMarker()
        result = translate("dummy", default=marker)
        self.assertTrue(result is marker)

    def test_msgid_translationstring_translator_is_None(self):
        msgid = DummyTranslationString('abc')
        translate = self._makeOne(None)
        result = translate(msgid)
        self.assertEqual(result, 'interpolated')

    def test_msgid_text_type_translator_is_None(self):
        msgid = u('foo')
        translate = self._makeOne(None)
        result = translate(msgid)
        self.assertEqual(result, u('foo'))

    def test_msgid_translationstring_translator_is_not_None(self):
        msgid = DummyTranslationString()
        def translator(msg):
            return msg
        translate = self._makeOne(translator)
        result = translate(msgid)
        self.assertEqual(result, msgid)

    def test_msgid_text_type_translator_is_not_None(self):
        msgid = 'foo'
        def translator(msg):
            return msg
        translate = self._makeOne(translator)
        result = translate(msgid)
        self.assertEqual(result, msgid)


class TestTranslator(unittest.TestCase):
    def _makeOne(self, translations=None, policy=None):
        from translationstring import Translator
        return Translator(translations, policy)

    def test_translations_None_interpolation_required(self):
        inst = self._makeOne()
        tstring = DummyTranslationString('$abc', mapping=True)
        result = inst(tstring)
        self.assertEqual(result, 'interpolated')
        
    def test_translations_None_interpolation_not_required(self):
        inst = self._makeOne()
        tstring = DummyTranslationString('msgid', mapping=False)
        result = inst(tstring)
        self.assertEqual(result, 'msgid')

    def test_translate_normal_string_with_domain(self):
        translations = DummyTranslations('yo')
        inst = self._makeOne(translations)
        result = inst('abc', 'domain')
        self.assertEqual(result, 'yo')
        self.assertEqual(translations.asked_domain, 'domain')

    def test_translate_normal_string_with_no_domain(self):
        translations = DummyTranslations('yo')
        inst = self._makeOne(translations)
        result = inst('abc')
        self.assertEqual(result, 'yo')
        self.assertEqual(translations.asked_domain, 'messages')

    def test_translate_normal_string_with_mapping(self):
        inst = self._makeOne(None)
        result = inst('${a}', mapping={'a':1})
        self.assertEqual(result, '1')

    def test_policy_returns_msgid(self):
        tstring = DummyTranslationString('msgid', default='default')
        def policy(translations, msg, domain, context):
            return msg
        inst = self._makeOne('ignoreme', policy)
        result = inst(tstring)
        self.assertEqual(result, 'default')

    def test_policy_returns_translation(self):
        tstring = DummyTranslationString('msgid')
        def policy(translations, msg, domain, context):
            return 'translated'
        inst = self._makeOne('ignoreme', policy)
        result = inst(tstring)
        self.assertEqual(result, 'translated')

class TestPluralizer(unittest.TestCase):
    def _makeOne(self, translations=None, policy=None):
        from translationstring import Pluralizer
        return Pluralizer(translations, policy)

    def test_translations_None_interpolation_required(self):
        inst = self._makeOne()
        result = inst('$abc', '$abc', 1, mapping={'abc':1})
        self.assertEqual(result, '1')
        
    def test_translations_None_interpolation_not_required(self):
        inst = self._makeOne()
        result = inst('msgid', 'msgid', 1)
        self.assertEqual(result, 'msgid')

    def test_policy_returns_translated(self):
        translations = DummyTranslations('result')
        def policy(translations, singular, plural, n, domain, context):
            return 'translated'
        inst = self._makeOne(translations, policy)
        tstring = DummyTranslationString('msgid')
        result = inst(tstring, tstring, 1)
        self.assertEqual(result, 'translated')

class Test_ugettext_policy(unittest.TestCase):
    def _callFUT(self, translations, tstring, domain, context):
        from translationstring import ugettext_policy
        return ugettext_policy(translations, tstring, domain, context)

    def test_it(self):
        translations = DummyTranslations('result')
        result = self._callFUT(translations, 'string', None, None)
        self.assertEqual(result, 'result')

    def test_msgctxt(self):
        translations = DummyTranslations('result')
        result = self._callFUT(translations, u('p\xf8f'), None, 'button')
        self.assertEqual(translations.params, (u('button\x04p\xf8f'),))
        self.assertEqual(result, 'result')

    def test_msgctxt_no_translation_found(self):
        input = u('p\xf8f')
        translations = DummyTranslations(input)
        result = self._callFUT(translations, input, None, 'button')
        self.assertEqual(result, u('p\xf8f'))

class Test_dugettext_policy(unittest.TestCase):
    def _callFUT(self, translations, tstring, domain, context=None):
        from translationstring import dugettext_policy
        return dugettext_policy(translations, tstring, domain, context)

    def test_it_use_default_domain(self):
        translations = DummyTranslations('result', domain=None)
        tstring = DummyTranslationString()
        result = self._callFUT(translations, tstring, None)
        self.assertEqual(result, 'result')
        self.assertEqual(translations.asked_domain, 'messages')

    def test_it_use_translations_domain(self):
        translations = DummyTranslations('result', domain='notdefault')
        tstring = DummyTranslationString()
        result = self._callFUT(translations, tstring, None)
        self.assertEqual(result, 'result')
        self.assertEqual(translations.asked_domain, 'notdefault')

    def test_it_use_tstring_domain(self):
        translations = DummyTranslations('result', domain='notdefault')
        tstring = DummyTranslationString(domain='exact')
        result = self._callFUT(translations, tstring, None)
        self.assertEqual(result, 'result')
        self.assertEqual(translations.asked_domain, 'exact')

    def test_it_use_explicit_domain(self):
        translations = DummyTranslations('result')
        tstring = DummyTranslationString()
        result = self._callFUT(translations, tstring, 'yo')
        self.assertEqual(result, 'result')
        self.assertEqual(translations.asked_domain, 'yo')

    def test_it_translations_has_no_dugettext(self):
        translations = DummyTranslations('result', domain='foo')
        tstring = DummyTranslationString('abc')
        translations.dugettext = None
        result = self._callFUT(translations, tstring, None)
        self.assertEqual(result, 'result')

    def test_msgctxt_from_tstring(self):
        translations = DummyTranslations('result')
        tstring = DummyTranslationString(u('p\xf8f'), context='button')
        result = self._callFUT(translations, tstring, None)
        self.assertEqual(translations.params, ('messages', u('button\x04p\xf8f'),))
        self.assertEqual(result, 'result')

    def test_msgctxt_override(self):
        translations = DummyTranslations('result')
        tstring = DummyTranslationString(u('p\xf8f'), context='other')
        result = self._callFUT(translations, tstring, None, context='button')
        self.assertEqual(translations.params, ('messages', u('button\x04p\xf8f'),))
        self.assertEqual(result, 'result')

    def test_msgctxt_no_translation_found(self):
        translations = DummyTranslations(u('button\x04p\xf8f'))
        tstring = DummyTranslationString(u('p\xf8f'), context='button')
        result = self._callFUT(translations, tstring, None)
        self.assertEqual(result, u('p\xf8f'))

class Test_ungettext_policy(unittest.TestCase):
    def _callFUT(self, translations, singular, plural, n, domain=None,
                 mapping=None, context=None):
        from translationstring import ungettext_policy
        return ungettext_policy(translations, singular, plural, n, domain, context)

    def test_it(self):
        translations = DummyTranslations('result')
        result = self._callFUT(translations, 'singular', 'plural', 1)
        self.assertEqual(result, 'result')

    def test_msgctxt(self):
        translations = DummyTranslations('result')
        result = self._callFUT(translations, u('p\xf8f'), 'plural', 1, context='button')
        self.assertEqual(translations.params, (u('button\x04p\xf8f'), 'plural', 1))
        self.assertEqual(result, 'result')

    def test_msgctxt_no_translation(self):
        translations = DummyTranslations(u('button\x04p\xf8f'))
        result = self._callFUT(translations, u('p\xf8f'), 'plural', 1, context='button')
        self.assertEqual(translations.params, (u('button\x04p\xf8f'), 'plural', 1))
        self.assertEqual(result, u('p\xf8f'))

class Test_dungettext_policy(unittest.TestCase):
    def _callFUT(self, translations, singular, plural, n, domain=None,
                 mapping=None, context=None):
        from translationstring import dungettext_policy
        return dungettext_policy(translations, singular, plural, n, domain, context)

    def test_it_use_default_domain(self):
        translations = DummyTranslations('result')
        result = self._callFUT(translations, 'singular', 'plural', 1)
        self.assertEqual(result, 'result')
        self.assertEqual(translations.asked_domain, 'messages')

    def test_it_use_translation_domain(self):
        translations = DummyTranslations('result', domain='translation')
        result = self._callFUT(translations, 'singular', 'plural', 1, None)
        self.assertEqual(result, 'result')
        self.assertEqual(translations.asked_domain, 'translation')

    def test_it_use_passed_domain(self):
        translations = DummyTranslations('result', domain='translation')
        result = self._callFUT(translations, 'singular', 'plural', 1, 'domain')
        self.assertEqual(result, 'result')
        self.assertEqual(translations.asked_domain, 'domain')

    def test_it_translations_has_no_dungettext(self):
        translations = DummyTranslations('result', domain='translation')
        translations.dungettext = None
        result = self._callFUT(translations, 'singular', 'plural', 1, 'domain')
        self.assertEqual(result, 'result')

class DummyTranslations(object):
    def __init__(self, result, domain=None):
        self.result = result
        self.domain = domain
        
    def gettext(self, tstring): # pragma: no cover
        self.params = (tstring,)
        return self.result
    
    def ngettext(self, singular, plural, n): # pragma: no cover
        self.params = (singular, plural, n)
        return self.result

    def ugettext(self, tstring): # pragma: no cover
        self.params = (tstring,)
        return self.result

    def dugettext(self, domain, tstring): # pragma: no cover
        self.params = (domain, tstring)
        self.asked_domain = domain
        return self.result

    def ungettext(self, singular, plural, n): # pragma: no cover
        self.params = (singular, plural, n)
        return self.result

    def dungettext(self, domain, singular, plural, n): # pragma: no cover
        self.params = (domain, singular, plural, n)
        self.asked_domain = domain
        return self.result

class DummyTranslationString(text_type):
    def __new__(cls, msgid='', domain=None, default=None, mapping=None, context=None):
        self = text_type.__new__(cls, msgid)
        text_type.__init__(self, msgid)
        self.domain = domain
        self.context = context
        self.mapping = mapping
        if default is None:
            default = msgid
        self.default = default
        return self
        
    def interpolate(self, translated=None):
        return 'interpolated'
    



import re
from gettext import NullTranslations
from translationstring.compat import text_type
from translationstring.compat import string_types
from translationstring.compat import PY3

NAME_RE = r"[a-zA-Z][-a-zA-Z0-9_]*"

_interp_regex = re.compile(r'(?<!\$)(\$(?:(%(n)s)|{(%(n)s)}))'
    % ({'n': NAME_RE}))

CONTEXT_MASK = text_type('%s\x04%s')

class TranslationString(text_type):
    """
    The constructor for a :term:`translation string`.  A translation
    string is a Unicode-like object that has some extra metadata.

    This constructor accepts one required argument named ``msgid``.
    ``msgid`` must be the :term:`message identifier` for the
    translation string.  It must be a ``unicode`` object or a ``str``
    object encoded in the default system encoding.

    Optional keyword arguments to this object's constructor include
    ``domain``, ``default``, and ``mapping``.

    ``domain`` represents the :term:`translation domain`.  By default,
    the translation domain is ``None``, indicating that this
    translation string is associated with the default translation
    domain (usually ``messages``).

    ``default`` represents an explicit *default text* for this
    translation string.  Default text appears when the translation
    string cannot be translated.  Usually, the ``msgid`` of a
    translation string serves double duty as its default text.
    However, using this option you can provide a different default
    text for this translation string.  This feature is useful when the
    default of a translation string is too complicated or too long to
    be used as a message identifier. If ``default`` is provided, it
    must be a ``unicode`` object or a ``str`` object encoded in the
    default system encoding (usually means ASCII).  If ``default`` is
    ``None`` (its default value), the ``msgid`` value used by this
    translation string will be assumed to be the value of ``default``.

    ``mapping``, if supplied, must be a dictionary-like object which
    represents the replacement values for any :term:`translation
    string` *replacement marker* instances found within the ``msgid``
    (or ``default``) value of this translation string.

    ``context`` represents the :term:`translation context`.  By default,
    the translation context is ``None``.

    After a translation string is constructed, it behaves like most
    other ``unicode`` objects; its ``msgid`` value will be displayed
    when it is treated like a ``unicode`` object.  Only when its
    ``ugettext`` method is called will it be translated.

    Its default value is available as the ``default`` attribute of the
    object, its :term:`translation domain` is available as the
    ``domain`` attribute, and the ``mapping`` is available as the
    ``mapping`` attribute.  The object otherwise behaves much like a
    Unicode string.
    """
    __slots__ = ('domain', 'context', 'default', 'mapping')

    def __new__(self, msgid, domain=None, default=None, mapping=None, context=None):

        # NB: this function should never never lose the *original
        # identity* of a non-``None`` but empty ``default`` value
        # provided to it.  See the comment in ChameleonTranslate.

        self = text_type.__new__(self, msgid)
        if isinstance(msgid, self.__class__):
            domain = domain or msgid.domain and msgid.domain[:]
            context = context or msgid.context and msgid.context[:]
            default = default or msgid.default and msgid.default[:]
            mapping = mapping or msgid.mapping and msgid.mapping.copy()
            msgid = text_type(msgid)
        self.domain = domain
        self.context = context
        if default is None:
            default = text_type(msgid)
        self.default = default
        self.mapping = mapping
        return self

    def __mod__(self, options):
        """Create a new TranslationString instance with an updated mapping.
        This makes it possible to use the standard python %-style string
        formatting with translatable strings. Only dictionary
        arguments are supported.
        """
        if not isinstance(options, dict):
            raise ValueError(
                    'Can only interpolate translationstring '
                    'with dictionaries.')
        if self.mapping:
            mapping = self.mapping.copy()
            mapping.update(options)
        else:
            mapping = options.copy()
        return TranslationString(self, mapping=mapping)

    def interpolate(self, translated=None):
        """ Interpolate the value ``translated`` which is assumed to
        be a Unicode object containing zero or more *replacement
        markers* (``$foo`` or ``${bar}``) using the ``mapping``
        dictionary attached to this instance.  If the ``mapping``
        dictionary is empty or ``None``, no interpolation is
        performed.

        If ``translated`` is ``None``, interpolation will be performed
        against the ``default`` value.
        """
        if translated is None:
            translated = self.default

        # NB: this function should never never lose the *original
        # identity* of a non-``None`` but empty ``default`` value it
        # is provided.  If (translated == default) , it should return the
        # *original* default, not a derivation.  See the comment below in
        # ChameleonTranslate.

        if self.mapping and translated:
            def replace(match):
                whole, param1, param2 = match.groups()
                return text_type(self.mapping.get(param1 or param2, whole))
            translated = _interp_regex.sub(replace, translated)

        return translated

    def __reduce__(self):
        return self.__class__, self.__getstate__()

    def __getstate__(self):
        return text_type(self), self.domain, self.default, self.mapping, self.context

def TranslationStringFactory(factory_domain):
    """ Create a factory which will generate translation strings
    without requiring that each call to the factory be passed a
    ``domain`` value.  A single argument is passed to this class'
    constructor: ``domain``.  This value will be used as the
    ``domain`` values of :class:`translationstring.TranslationString`
    objects generated by the ``__call__`` of this class.  The
    ``msgid``, ``mapping``, and ``default`` values provided to the
    ``__call__`` method of an instance of this class have the meaning
    as described by the constructor of the
    :class:`translationstring.TranslationString`"""
    def create(msgid, mapping=None, default=None, context=None):
        """ Provided a msgid (Unicode object or :term:`translation
        string`) and optionally a mapping object, and a *default
        value*, return a :term:`translation string` object."""

        # if we are passing in a TranslationString as the msgid, then
        # use its domain
        if isinstance(msgid, TranslationString):
            domain = msgid.domain or factory_domain
        else:
            domain = factory_domain

        return TranslationString(msgid, domain=domain, default=default,
                                 mapping=mapping, context=context)
    return create

def ChameleonTranslate(translator):
    """
    When necessary, use the result of calling this function as a
    Chameleon template 'translate' function (e.g. the ``translate``
    argument to the ``chameleon.zpt.template.PageTemplate``
    constructor) to allow our translation machinery to drive template
    translation.  A single required argument ``translator`` is
    passsed.  The ``translator`` provided should be a callable which
    accepts a single argument ``translation_string`` ( a
    :class:`translationstring.TranslationString` instance) which
    returns a ``unicode`` object as a translation.  ``translator`` may
    also optionally be ``None``, in which case no translation is
    performed (the ``msgid`` or ``default`` value is returned
    untranslated).
    """
    def translate(msgid, domain=None, mapping=None, context=None,
                 target_language=None, default=None):

        # NB: note that both TranslationString._init__ and
        # TranslationString.interpolate are careful to never lose the
        # *identity* of an empty but non-``None`` ``default`` value we
        # provide to them.  For example, neither of those functions
        # are permitted to run an empty but non-``None`` ``default``
        # through ``unicode`` and throw the original default value
        # away afterwards.

        # This has a dubious cause: for Chameleon API reasons we must
        # ensure that, after translation, if ( (translated == msgid)
        # and (not default) and (default is not None) ) that we return
        # the ``default`` value provided to us *unmodified*, because
        # Chameleon uses it as a sentinel (it compares the return
        # value of this function by identity to what it passed in as
        # ``default``; this marker is a
        # chameleon.core.i18n.StringMarker instance, a subclass of str
        # that == '').  This is, of course, totally absurd, because
        # Chameleon *also* wants us to use ``default`` as the input to
        # a translation string in some cases, and maintaining the
        # identity of this object through translation operations isn't
        # a contract it spells out in its docs.

        # Chameleon's use of ``default`` to represent both a sentinel
        # and input to a translation string is a Chameleon i18n
        # extensibility design bug.  Until we overhaul its hook point
        # for translation extensibility, we need to appease it by
        # preserving ``default`` in the aforementioned case.  So we
        # spray these indignant comments all over this module. ;-)

        if not isinstance(msgid, string_types):
            if msgid is not None:
                msgid = text_type(msgid)
            return msgid

        tstring = msgid

        if not hasattr(tstring, 'interpolate'):
            tstring = TranslationString(msgid, domain, default, mapping, context)
        if translator is None:
            result = tstring.interpolate()
        else:
            result = translator(tstring)

        return result

    return translate

def ugettext_policy(translations, tstring, domain, context):
    """ A translator policy function which unconditionally uses the
    ``ugettext`` API on the translations object."""

    if PY3: # pragma: no cover
        _gettext = translations.gettext
    else: # pragma: no cover
        _gettext = translations.ugettext

    if context:
	# Workaround for http://bugs.python.org/issue2504?
        msgid = CONTEXT_MASK % (context, tstring)
    else:
        msgid = tstring

    translated = _gettext(msgid)
    return tstring if translated == msgid else translated

def dugettext_policy(translations, tstring, domain, context):
    """ A translator policy function which assumes the use of a
    :class:`babel.support.Translations` translations object, which
    supports the dugettext API; fall back to ugettext."""
    if domain is None:
        default_domain = getattr(translations, 'domain', None) or 'messages'
        domain = getattr(tstring, 'domain', None) or default_domain
    context = context or getattr(tstring, 'context', None)
    if context:
	# Workaround for http://bugs.python.org/issue2504?
        msgid = CONTEXT_MASK % (context, tstring)
    else:
        msgid = tstring

    if getattr(translations, 'dugettext', None) is not None:
        translated = translations.dugettext(domain, msgid)
    else:
        if PY3: # pragma: no cover
            _gettext = translations.gettext
        else: # pragma: no cover
            _gettext = translations.ugettext

        translated = _gettext(msgid)
    return tstring if translated == msgid else translated

def Translator(translations=None, policy=None):
    """
    Return a translator object based on the ``translations`` and
    ``policy`` provided.  ``translations`` should be an object
    supporting *at least* the Python :class:`gettext.NullTranslations`
    API but ideally the :class:`babel.support.Translations` API, which
    has support for domain lookups like dugettext.

    ``policy`` should be a callable which accepts three arguments:
    ``translations``, ``tstring`` and ``domain``.  It must perform the
    actual translation lookup.  If ``policy`` is ``None``, the
    :func:`translationstring.dugettext_policy` policy will be used.

    The callable returned accepts three arguments: ``tstring``
    (required), ``domain`` (optional) and ``mapping`` (optional).
    When called, it will translate the ``tstring`` translation string
    to a ``unicode`` object using the ``translations`` provided.  If
    ``translations`` is ``None``, the result of interpolation of the
    default value is returned.  The optional ``domain`` argument can
    be used to specify or override the domain of the ``tstring``
    (useful when ``tstring`` is a normal string rather than a
    translation string).  The optional ``mapping`` argument can
    specify or override the ``tstring`` interpolation mapping, useful
    when the ``tstring`` argument is a simple string instead of a
    translation string.
    """
    if policy is None:
        policy = dugettext_policy
    def translator(tstring, domain=None, mapping=None, context=None):
        if not hasattr(tstring, 'interpolate'):
            tstring = TranslationString(tstring, domain=domain, mapping=mapping, context=context)
        elif mapping:
            if tstring.mapping:
                new_mapping = tstring.mapping.copy()
                new_mapping.update(mapping)
            else:
                new_mapping = mapping
            tstring = TranslationString(tstring, domain=domain, mapping=new_mapping, context=context)
        translated = tstring
        domain = domain or tstring.domain
        context = context or tstring.context
        if translations is not None:
            translated = policy(translations, tstring, domain, context)
        if translated == tstring:
            translated = tstring.default
        if translated and '$' in translated and tstring.mapping:
            translated = tstring.interpolate(translated)
        return translated
    return translator

def ungettext_policy(translations, singular, plural, n, domain, context):
    """ A pluralizer policy function which unconditionally uses the
    ``ungettext`` API on the translations object."""

    if PY3: # pragma: no cover
        _gettext = translations.ngettext
    else: # pragma: no cover
        _gettext = translations.ungettext

    if context:
	# Workaround for http://bugs.python.org/issue2504?
        msgid = CONTEXT_MASK % (context, singular)
    else:
        msgid = singular

    translated = _gettext(msgid, plural, n)
    return singular if translated == msgid else translated

def dungettext_policy(translations, singular, plural, n, domain, context):
    """ A pluralizer policy function which assumes the use of the
    :class:`babel.support.Translations` class, which supports the
    dungettext API; falls back to ungettext."""

    default_domain = getattr(translations, 'domain', None) or 'messages'
    domain = domain or default_domain
    if context:
	# Workaround for http://bugs.python.org/issue2504?
        msgid = CONTEXT_MASK % (context, singular)
    else:
        msgid = singular
    if getattr(translations, 'dungettext', None) is not None:
        translated = translations.dungettext(domain, msgid, plural, n)
    else:
        if PY3: # pragma: no cover
            _gettext = translations.ngettext
        else: # pragma: no cover
            _gettext = translations.ungettext

        translated = _gettext(msgid, plural, n)
    return singular if translated == msgid else translated

def Pluralizer(translations=None, policy=None):
    """
    Return a pluralizer object based on the ``translations`` and
    ``policy`` provided.  ``translations`` should be an object
    supporting *at least* the Python :class:`gettext.NullTranslations`
    API but ideally the :class:`babel.support.Translations` API, which
    has support for domain lookups like dugettext.

    ``policy`` should be a callable which accepts five arguments:
    ``translations``, ``singular`` and ``plural``, ``n`` and
    ``domain``.  It must perform the actual pluralization lookup.  If
    ``policy`` is ``None``, the
    :func:`translationstring.dungettext_policy` policy will be used.

    The object returned will be a callable which has the following
    signature::

        def pluralizer(singular, plural, n, domain=None, mapping=None):
            ...

    The ``singular`` and ``plural`` objects passed may be translation
    strings or unicode strings.  ``n`` represents the number of
    elements.  ``domain`` is the translation domain to use to do the
    pluralization, and ``mapping`` is the interpolation mapping that
    should be used on the result.  Note that if the objects passed are
    translation strings, their domains and mappings are ignored.  The
    domain and mapping arguments must be used instead.  If the ``domain`` is
    not supplied, a default domain is used (usually ``messages``).
    """

    if policy is None:
        policy = dungettext_policy
    if translations is None:
        translations = NullTranslations()
    def pluralizer(singular, plural, n, domain=None, mapping=None, context=None):
        """ Pluralize this object """
        translated = text_type(
            policy(translations, singular, plural, n, domain, context))
        if translated and '$' in translated and mapping:
            return TranslationString(translated, mapping=mapping).interpolate()
        return translated
    return pluralizer



import sys

PY3 = sys.version_info[0] == 3

if PY3: # pragma: no cover
    string_types = (str,)
    text_type = str
else: # pragma: no cover
    string_types = (basestring,)
    text_type = unicode

if PY3: # pragma: no cover
    def u(s):
        return s
else: # pragma: no cover
    def u(s):
        return unicode(s, 'unicode_escape')




import gettext
from translationstring.compat import PY3

class Translations(gettext.GNUTranslations, object):
    """An extended translation catalog class."""

    DEFAULT_DOMAIN = 'messages'

    def __init__(self, fileobj=None, domain=DEFAULT_DOMAIN):
        """Initialize the translations catalog.

        :param fileobj: the file-like object the translation should be read
                        from
        """
        gettext.GNUTranslations.__init__(self, fp=fileobj)
        self.files = filter(None, [getattr(fileobj, 'name', None)])
        self.domain = domain
        self._domains = {}
        if fileobj is not None:
            fileobj.close()

    def load(cls, dirname=None, locales=None, domain=DEFAULT_DOMAIN):
        """Load translations from the given directory.

        :param dirname: the directory containing the ``MO`` files
        :param locales: the list of locales in order of preference (items in
                        this list can be either `Locale` objects or locale
                        strings)
        :param domain: the message domain
        :return: the loaded catalog, or a ``NullTranslations`` instance if no
                 matching translations were found
        :rtype: `Translations`
        """
        if locales is not None:
            locales = [str(locale) for locale in locales]
        filename = gettext.find(domain, dirname, locales)
        fp = open(filename, 'rb')
        return cls(fileobj=fp, domain=domain)
    load = classmethod(load)
    
    def dugettext(self, domain, message):
        """Like ``ugettext()``, but look the message up in the specified
        domain.
        """
        if PY3: # pragma: no cover
            return self._domains.get(domain, self).gettext(message)
        else: # pragma: no cover
            return self._domains.get(domain, self).ugettext(message)
    
    def dungettext(self, domain, singular, plural, num):
        """Like ``ungettext()`` but look the message up in the specified
        domain.
        """
        if PY3: # pragma: no cover
            return self._domains.get(domain, self).ngettext(
                singular, plural, num)
        else: # pragma: no cover
            return self._domains.get(domain, self).ungettext(
                singular, plural, num)
        
    # Most of the downwards code, until it get's included in stdlib, from:
    #    http://bugs.python.org/file10036/gettext-pgettext.patch
    #    
    # The encoding of a msgctxt and a msgid in a .mo file is
    # msgctxt + "\x04" + msgid (gettext version >= 0.15)
    CONTEXT_ENCODING = '%s\x04%s'
